/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.messaging.exchange;

import java.io.Serializable;
import java.net.URI;
import java.util.Set;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

/**
 * 
 * This class wrappers a real <code>MessageExchange</code>.
 * <p>
 * Each actor (consumer or provider) owns an instance of this class. <br>
 * The wrapper checks that the methods which are called by the related component can be used at the current exchange
 * state.
 * </p>
 * 
 * <p>
 * This class is also needed so that the same object is used for the answer of a
 * {@link DeliveryChannel#sendSync(MessageExchange)}: for example with a remote transporter, the answer will be another
 * message exchange and it will take the place of the {@link #messageExchange} in this wrapper.
 * </p>
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Victor Noël - Linagora
 */
public class MessageExchangeWrapper implements MessageExchange, Serializable {

    private static final long serialVersionUID = 2388703250533525122L;

    /**
     * The wrapped Message Exchange.
     * 
     * It can't be null, but it can be replaced, for example when we get an answer from a sendSync.
     */
    protected PetalsMessageExchange messageExchange;

    /**
     * <code>true</code> if the owner of the object has ownership on {@link #messageExchange}.
     */
    private boolean hasOwnership;

    /**
     * Create a {@link MessageExchangeWrapper} that has ownership over the {@link MessageExchange} that it contains.
     * 
     */
    public MessageExchangeWrapper(final PetalsMessageExchange me) {
        assert me != null;
        this.messageExchange = me;
        this.hasOwnership = true;
    }
    
    @Override
    public Fault createFault() throws MessagingException {
        this.checkComponentOwnsExchange();

        return this.messageExchange.createFault();
    }

    @Override
    public NormalizedMessage createMessage() throws MessagingException {
        this.checkComponentOwnsExchange();

        return this.messageExchange.createMessage();
    }

    @Override
    public ServiceEndpoint getEndpoint() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getEndpoint();
    }

    @Override
    public Exception getError() {
        checkComponentOwnsExchange();

        return this.messageExchange.getError();
    }

    @Override
    public String getExchangeId() {
        return this.messageExchange.getExchangeId();
    }
    
    @Override
    public Fault getFault() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getFault();
    }

    @Override
    public QName getInterfaceName() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getInterfaceName();
    }

    @Override
    public NormalizedMessage getMessage(final String name) {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getMessage(name);
    }

    @Override
    public QName getOperation() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getOperation();
    }

    @Override
    public URI getPattern() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getPattern();
    }

    @Override
    public Object getProperty(final String name) {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getProperty(name);
    }

    @Override
    public Set<?> getPropertyNames() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getPropertyNames();
    }

    @Override
    public Role getRole() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getRole();
    }

    @Override
    public QName getService() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.getService();
    }

    @Override
    public ExchangeStatus getStatus() {
        return this.messageExchange.getStatus();
    }

    @Override
    public boolean isTransacted() {
        this.checkComponentOwnsExchange();

        return this.messageExchange.isTransacted();
    }

    @Override
    public void setEndpoint(final ServiceEndpoint endpoint) {
        this.checkComponentOwnsExchange();

        this.messageExchange.setEndpoint(endpoint);
    }

    @Override
    public void setError(final Exception error) {
        this.checkComponentOwnsExchange();

        this.messageExchange.setError(error);
    }

    @Override
    public void setFault(final Fault fault) throws MessagingException {
        this.checkComponentOwnsExchange();

        this.messageExchange.setFault(fault);
    }

    @Override
    public void setInterfaceName(final QName interfaceName) {
        this.checkComponentOwnsExchange();

        this.messageExchange.setInterfaceName(interfaceName);
    }

    @Override
    public void setMessage(final NormalizedMessage msg, final String name) throws MessagingException {
        this.checkComponentOwnsExchange();

        this.messageExchange.setMessage(msg, name);
    }

    @Override
    public void setOperation(final QName name) {
        this.checkComponentOwnsExchange();

        this.messageExchange.setOperation(name);
    }

    @Override
    public void setProperty(final String name, final Object obj) {
        this.checkComponentOwnsExchange();

        this.messageExchange.setProperty(name, obj);
    }

    @Override
    public void setService(final QName service) {
        this.checkComponentOwnsExchange();

        this.messageExchange.setService(service);
    }

    @Override
    public void setStatus(final ExchangeStatus status) throws MessagingException {
        this.checkComponentOwnsExchange();

        this.messageExchange.setStatus(status);
    }

    /**
     * Return the real implementation of this {@link MessageExchange} and remove owneship over the exchange
     */
    public PetalsMessageExchange removeOwnership() {
        assert this.hasOwnership;
        this.hasOwnership = false;
        return this.messageExchange;
    }

    /**
     * Set the new {@link MessageExchange} and return ownership over it
     */
    public void returnOwnership(final PetalsMessageExchange exchange) {
        assert !this.hasOwnership;
        this.hasOwnership = true;
        this.messageExchange = exchange;
    }

    /**
     * @deprecated TODO this must be removed: for now it is used by the jbi-adapter-impl but it's really strange what
     *             they do with synchonized exchange... there must be bugs there for sure!!!
     */
    @Deprecated
    public PetalsMessageExchange getMessageExchange() {
        return this.messageExchange;
    }

    /**
     * Check that the component owns the message exchange.
     * 
     * @throws IllegalStateException
     */
    private void checkComponentOwnsExchange() {
        if (!this.hasOwnership) {
            throw new IllegalStateException("The component does not owns the MessageExchange");
        }
    }
}
