/**
 * Copyright (c) 2005-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.jbi.messaging.exchange.impl;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.security.auth.Subject;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.xml.BytesSource;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;

import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.SourceHelper;

/**
 * JBI <code>NormalizedMessage</code> implementation.
 * 
 * It ensures that the content is stored as a Source readable multiple time, i.e. that does not need forking!
 * 
 * @since Petals 1.0
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Victor Noël - Linagora
 */
public class NormalizedMessageImpl implements NormalizedMessage, Serializable {

    /**
     * The message exchange property name to set the compression of the exchange
     * payload.
     */
    public final static String PROPERTY_COMPRESSION = "org.ow2.petals.transport.compress";

    /**
     * The unique serial UID.
     */
    private static final long serialVersionUID = 45438732L;

    /**
     * Attachments to the main message: id + content.
     */
    protected transient Map<String, DataHandler> attachments = new HashMap<String, DataHandler>();

    /**
     * The XML "payload" message.
     */
    protected transient Source content;

    /**
     * Metadata associated with the message content. Map: property + value.
     */
    protected final Map<String, Object> properties = new HashMap<String, Object>();

    /**
     * JAAS Security Subject.
     */
    protected Subject subject;

    @Override
    public void addAttachment(final String id, final DataHandler attachment) throws MessagingException {
        attachments.put(id, attachment);
    }

    @Override
    public DataHandler getAttachment(final String id) {
        return attachments.get(id);
    }

    @Override
    public Set<?> getAttachmentNames() {
        return attachments.keySet();
    }

    @Override
    public Source getContent() {
        return content;
    }

    @Override
    public Object getProperty(final String name) {
        return properties.get(name);
    }

    @Override
    public Set<?> getPropertyNames() {
        return properties.keySet();
    }

    @Override
    public Subject getSecuritySubject() {
        return subject;
    }

    @Override
    public void removeAttachment(final String id) throws MessagingException {
        final Object att = attachments.remove(id);

        if (att == null) {
            throw new MessagingException(id + " attachment does not exist.");
        }
    }

    /**
     * This implementation will store the Source as a {@link DOMSource} if it's already one, or as a {@link BytesSource}
     * in the other cases.
     * 
     * Also, if it's a {@link DOMSource} and its content is a NodeImpl from the Java library, then it is marked as
     * read-only to prevent modification of shared in-memory messages by multiple actors.
     */
    @Override
    public void setContent(final Source newContent) throws MessagingException {
        if (isReReadable(newContent)) {
            this.content = makeReadOnly(newContent);
        } else if (newContent != null) {
            this.content = toBytesSource(newContent);
        } else {
            this.content = null;
        }
    }

    /**
     * These are the types we accept, if they conforms to some checks.
     * 
     */
    private boolean isReReadable(final Source source) {
        if (source instanceof BytesSource || source instanceof DOMSource) {
            return true;
        } else if (source instanceof StreamSource) {
            // if they are both null, then either it is an empty source or systemId gives the file to read
            // and it's then readable multiple times
            final StreamSource streamSource = (StreamSource) source;
            return streamSource.getInputStream() == null && streamSource.getReader() == null;
        } else if (source instanceof SAXSource) {
            final InputSource inputSource = ((SAXSource) source).getInputSource();
            // if inputSource is null then it is an empty source
            // if its streams are null then either it is an empty source or systemId gives the file to read and it is
            // readable multiple times
            return inputSource == null
                    || (inputSource.getByteStream() == null && inputSource.getCharacterStream() == null);
        } else {
            return false;
        }
    }

    /**
     * It is applied only to the types we accept and makes them readonly when possible.
     */
    private Source makeReadOnly(final Source source) throws MessagingException {

        final Source readOnlySource;
        if (source instanceof DOMSource) {
            Node node = ((DOMSource) source).getNode();

            if (node instanceof com.sun.org.apache.xerces.internal.impl.xs.opti.NodeImpl) {
                ((com.sun.org.apache.xerces.internal.impl.xs.opti.NodeImpl) node).setReadOnly(true, true);
            } else if (node instanceof com.sun.org.apache.xerces.internal.dom.NodeImpl) {
                ((com.sun.org.apache.xerces.internal.dom.NodeImpl) node).setReadOnly(true, true);
            } else {
                final Document doc = DocumentBuilders.newDocument();
                if (node instanceof Document) {
                    node = doc.importNode(((Document) node).getDocumentElement(), true);
                } else {
                    node = doc.importNode(node, true);
                }
                ((com.sun.org.apache.xerces.internal.dom.NodeImpl) node).setReadOnly(true, true);
            }
            readOnlySource = source;
        } else {
            // for now we don't have ways to make StreamSource or SAXSource unmodifiable...
            readOnlySource = source;
        }
        
        return readOnlySource;
    }

    private Source toBytesSource(final Source source) throws MessagingException {
        try {
            return new BytesSource(SourceHelper.toByteArray(source));
        } catch (final TransformerException e) {
            throw new MessagingException("Can't convert to BytesSource", e);
        }
    }

    @Override
    public void setProperty(final String name, final Object value) {
        if (value == null) {
            // case to remove properties not defined (null)
            properties.remove(name);
        } else {
            // define this property name with a new value
            properties.put(name, value);
        }
    }

    @Override
    public void setSecuritySubject(final Subject securitySubject) {
        this.subject = securitySubject;
    }

    /**
     * Used to deserialize Content object
     * 
     * @param s
     * @throws IOException
     */
    private void readObject(final ObjectInputStream s) throws IOException {
        try {
            s.defaultReadObject();

            attachments = MessageExchangeSerializer.deserializeAttachments(s);
            final String compression = (String) getProperty(PROPERTY_COMPRESSION);
            content = MessageExchangeSerializer.deserializeContent(s, Boolean.valueOf(compression));
        } catch (ClassNotFoundException e) {
            throw new IOException(e);
        }
    }

    /**
     * <p>
     * Serialization process :
     * <ul>
     * <li>Serialization of all serializable fields of the NormalizedMessage</li>
     * <li>Serialization of all attachment</li>
     * <li>Finally, serialization of the content.</li>
     * </ul>
     * </p>
     * <p>
     * <b>Note</b>: Content must always be serialized at the end of the
     * serialization process. Because in the deserialization process, a new
     * content will be created with all ending bytes of the Stream.
     * </p>
     * 
     * @param s
     * @throws IOException
     */
    private void writeObject(final ObjectOutputStream s) throws IOException {
        s.defaultWriteObject();
        try {
            MessageExchangeSerializer.serializeAttachments(attachments, s);
            final String compression = (String) getProperty(PROPERTY_COMPRESSION);
            MessageExchangeSerializer.serializeContent(content, s, Boolean
                    .parseBoolean(compression));
        } catch (final Exception e) {
            throw new IOException(e);
        }
    }
}
